// SPDX-License-Identifier: AGPL-3.0-or-later
// SPDX-FileCopyrightText: 2023 swt23w23
package catering.staff;

import static org.salespointframework.core.Currencies.EURO;

import java.time.YearMonth;
import java.util.Optional;

import org.javamoney.moneta.Money;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.stereotype.Controller;
import org.springframework.ui.Model;
import org.springframework.validation.Errors;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.ModelAttribute;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestParam;

import jakarta.validation.Valid;

@Controller
public class StaffController {

	private final StaffManagement staffManagement;

	public StaffController(StaffManagement staffManagement) {
		this.staffManagement = staffManagement;
	}

	@GetMapping("/staff")
	@PreAuthorize("hasRole('ADMIN')")
	public String getStaff(Model model, @RequestParam Optional<YearMonth> month) {
		return getStaff(model, new StaffForm(), month.orElseGet(YearMonth::now));
	}

	public String getStaff(Model model, @Valid StaffForm form, YearMonth month) {
		model.addAttribute("staff", staffManagement.findAll());
		model.addAttribute("form", form);
		model.addAttribute("month", month);
		model.addAttribute("management", staffManagement);
		return "staff";
	}

	@PostMapping("/staff/add")
	@PreAuthorize("hasRole('ADMIN')")
	public String addEmployee(@Valid @ModelAttribute("form") StaffForm form, Errors result, Model model) {
		form.validate(result);
		if (result.hasErrors()) {
			return getStaff(model, form, YearMonth.now());
		}
		staffManagement.save(new Employee(form.getName(), form.getJob(), Money.of(form.getWage(), EURO)));
		return "redirect:/staff";
	}

	@PostMapping("/staff/remove")
	@PreAuthorize("hasRole('ADMIN')")
	public String removeEmployee(@RequestParam("id") Employee employee, Model model) {
		staffManagement.delete(employee.getId());
		return "redirect:/staff";
	}

	@GetMapping("/staff/edit/{id}")
	@PreAuthorize("hasRole('ADMIN')")
	public String editEmployee(@PathVariable("id") Employee employee, Model model) {
		StaffForm form = new StaffForm();
		form.setJob(employee.getJob());
		form.setName(employee.getName());
		form.setWage(employee.getWage());
		return editEmployee(model, form);
	}

	public String editEmployee(Model model, @Valid StaffForm form) {
		model.addAttribute("form", form);
		return "edit-staff";
	}

	@PostMapping("/staff/edit/{id}")
	@PreAuthorize("hasRole('ADMIN')")
	public String updateEmployee(@PathVariable("id") Employee employee, @Valid @ModelAttribute("form") StaffForm form,
			Errors result, Model model) {
		form.validate(result);
		if (result.hasErrors()) {
			return editEmployee(model, form);
		}
		employee.setJob(form.getJob());
		employee.setName(form.getName());
		employee.setWage(form.getWage());
		staffManagement.save(employee);
		return "redirect:/staff";
	}
}
