// SPDX-License-Identifier: AGPL-3.0-or-later
// SPDX-FileCopyrightText: 2023 swt23w23
package catering.users;

import org.springframework.validation.ValidationUtils;

import java.util.Optional;

import org.springframework.validation.Errors;

/**
 * An abstract class to hold form input for {@link User} data.
 *
 * It can be extended to validate input for different use cases.
 */
public abstract class UserForm {

	private final String username, address, fullName;

	public UserForm(String username, String address, String fullName) {
		this.username = username;
		this.address = address;
		this.fullName = fullName;
	}

	public String getUsername() {
		return username;
	}

	public String getAddress() {
		return address;
	}

	public String getFullName() {
		return fullName;
	}

	/**
	 * An extension of {@link UserForm} with a password attribute.
	 */
	public abstract static class UserFormWithPassword extends UserForm {
		private final String password;

		public UserFormWithPassword(String username, String address, String fullName, String password) {
			super(username, address, fullName);
			this.password = password;
		}

		public String getPassword() {
			return password;
		}

		/**
		 * Template Method for validating the password.
		 *
		 * @param e {@link Errors} from {@link FormValidator}
		 */
		public void validatePassword(Errors e) {
			validatePasswordGeneric(e);
			Optional.ofNullable(e.getFieldValue("password")).map(s -> (String) s).ifPresent(s -> {
				if (s.chars().anyMatch(Character::isISOControl)) {
					e.rejectValue("password", "password must only contain printable characters");
				}
			});
		}

		/**
		 * Primitive Operation for validating the password depending on the subclass.
		 *
		 * @param e {@link Errors} from {@link FormValidator}
		 */
		public abstract void validatePasswordGeneric(Errors e);

		/**
		 * An extension of {@link UserFormWithPassword} to validate input during registration.
		 *
		 * It requires the password to be not empty.
		 */
		public static class RegistrationForm extends UserFormWithPassword {
			RegistrationForm(String username, String address, String fullName, String password) {
				super(username, address, fullName, password);
			}

			public static RegistrationForm empty() {
				return new UserFormWithPassword.RegistrationForm("", "", "", "");
			}

			public void validatePasswordGeneric(Errors e) {
				ValidationUtils.rejectIfEmptyOrWhitespace(e, "password", "password must not be empty");
			}
		}

		/**
		 * An extension of {@link UserFormWithPassword} to validate input during profile editing.
		 *
		 * The password can be empty.
		 */
		public static class ProfileForm extends UserFormWithPassword {
			ProfileForm(String username, String address, String fullName, String password) {
				super(username, address, fullName, password);
			}

			public void validatePasswordGeneric(Errors e) {
				if (e.getFieldValue("password") == null) {
					e.rejectValue("password", "password must not be null");
				}
			}
		}
	}

	/**
	 * An extension of {@link UserForm} to validate input during customer profile editing.
	 */
	public static class AdminForm extends UserForm {
		AdminForm(String username, String address, String fullName) {
			super(username, address, fullName);
		}
	}
}
